/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "com_myriad_tmap_compute_Pedigree.h"
#include "ped.h"
#include "map.h"
#include "input.h"
#include "inputOrder.h"
#include "mapLlhd.h"
#include "improve.h"
#include "twoPoint.h"
#include "build.h"
#include <jni.h>
#include <stdlib.h>
#include <string.h>

#define MAKE_MAP_COMPUTE_DISTS \
                com_myriad_tmap_compute_Pedigree_MAKE_MAP_COMPUTE_DISTS
#define MAKE_MAP_BUILD com_myriad_tmap_compute_Pedigree_MAKE_MAP_BUILD
#define MAKE_MAP_IMPROVE com_myriad_tmap_compute_Pedigree_MAKE_MAP_IMPROVE

static int readPedFileJstring(JNIEnv *env, pedigree_t *ped, jstring path);
static void orderFromJStrings(JNIEnv *env, pedigree_t *ped, map_t *map,
                              jobjectArray markerNames, jboolean split);
static void exportMap(JNIEnv *env, map_t *map, jdoubleArray distsMoth,
                      jdoubleArray distsFath, jdoubleArray errorsMoth,
                      jdoubleArray errorsFath);
static int exportOrder(JNIEnv *env, jobjectArray markerNames, map_t *map);

JNIEXPORT jlong JNICALL
Java_com_myriad_tmap_compute_Pedigree_readPedigreeFile(JNIEnv *env,
                                                         jobject this,
                                                         jstring pedPath)
{
    pedigree_t  *ped = (pedigree_t *)malloc(sizeof(pedigree_t));
    
    if (readPedFileJstring(env, ped, pedPath))
        return (jlong)ped;
    else {
        free(ped);
        return 0;
    }
}

JNIEXPORT void JNICALL
Java_com_myriad_tmap_compute_Pedigree_freePed(JNIEnv *env, jobject this,
                                                jlong pedPtr)
{
    pedigree_t  *ped = (pedigree_t *)pedPtr;
    
    freePed(ped);
    free(ped);
}

JNIEXPORT jint JNICALL
Java_com_myriad_tmap_compute_Pedigree_getNumMarkers(JNIEnv *env, jobject this,
                                                      jlong pedPtr)
{
    return ((pedigree_t *)pedPtr)->numMarks;
}

JNIEXPORT void JNICALL
Java_com_myriad_tmap_compute_Pedigree_compute2PtsNative(JNIEnv *env,
                                                          jobject this,
                                                          jlong pedPtr,
                                                          jobjectArray names,
                                                          jobjectArray lods,
                                                          jobjectArray dists)
{
    pedigree_t  *ped = (pedigree_t *)pedPtr;
    double      *rowLods = (double *)malloc(ped->numMarks * sizeof(double));
    double      *rowDists = (double *)malloc(ped->numMarks * sizeof(double));
    int         i, j;
    
    for (i = 0; i < ped->numMarks; ++i) {
        jdoubleArray    jrowLods;
        jdoubleArray    jrowDists;
        jstring         str = (*env)->NewStringUTF(env, ped->markers[i].name);
        
        (*env)->SetObjectArrayElement(env, names, i, str);
        
        for (j = 0; j < i; ++j) {
            double  theta;
            theta = compute2PtPhased(ped->markers + i, ped->markers + j,
                                     ped->numInds, rowLods + j);
            rowDists[j] = theta2Dist(theta);
        }
            
        jrowLods = (*env)->NewDoubleArray(env, i);
        (*env)->SetDoubleArrayRegion(env, jrowLods, 0, i, rowLods);
        (*env)->SetObjectArrayElement(env, lods, i, jrowLods);
        
        jrowDists = (*env)->NewDoubleArray(env, i);
        (*env)->SetDoubleArrayRegion(env, jrowDists, 0, i, rowDists);
        (*env)->SetObjectArrayElement(env, dists, i, jrowDists);
    }
    
    free(rowLods);
    free(rowDists);
}

JNIEXPORT jobjectArray JNICALL
Java_com_myriad_tmap_compute_Pedigree_getOrderNative(JNIEnv *env,
                                                       jobject this,
                                                       jlong pedPtr,
                                                       jstring orderPath)
{
    const char      *cOrderPath;
    pedigree_t      *ped = (pedigree_t *)pedPtr;
    map_t           map;
    jclass          stringClass;
    jobjectArray    result;
    
    cOrderPath = (*env)->GetStringUTFChars(env, orderPath, NULL);
    map.order = readOrderFile(ped, &map.numMarks, cOrderPath);
    (*env)->ReleaseStringUTFChars(env, orderPath, cOrderPath);
    if(map.order == NULL) {
        fprintf(stderr, "Couldn't read order file\n");
        return NULL;
    }
    
    stringClass = (*env)->FindClass(env, "java/lang/String");
    if (stringClass == NULL) {
        freeMap(&map);
        return NULL;
    }
    result = (*env)->NewObjectArray(env, map.numMarks, stringClass, NULL);
    
    if (!exportOrder(env, result, &map))
        result = NULL;
    
    free(map.order);
    
    return result;
}

JNIEXPORT void JNICALL
Java_com_myriad_tmap_compute_Pedigree_makeMapNative(JNIEnv *env, jobject this,
                                                      jlong pedPtr,
                                                      jobjectArray markerNames,
                                                      jboolean split,
                                                      jdoubleArray distsMoth,
                                                      jdoubleArray distsFath,
                                                      jdoubleArray errorsMoth,
                                                      jdoubleArray errorsFath,
                                                      jint type)
{
    pedigree_t      *ped = (pedigree_t *)pedPtr;
    map_t           map;
    
    orderFromJStrings(env, ped, &map, markerNames, split);
    
    switch (type) {
        case MAKE_MAP_COMPUTE_DISTS:
            computeMapLlhd(&map);
            normalizeMapDists(&map);
            break;
        case MAKE_MAP_BUILD:
            {
                map_t   tmpMap;
                
                buildMap(&tmpMap, ped, map.numMarks, map.order, split);
                swapMaps(&map, &tmpMap);
                freeMap(&tmpMap);
                normalizeMap(&map);
            }
            break;
        case MAKE_MAP_IMPROVE:
            computeMapLlhd(&map);
            improveMap(&map);
            normalizeMap(&map);
            break;
        default:
            fprintf(stderr, "Bad make map type\n");
            break;
    }
    
    if (type == MAKE_MAP_COMPUTE_DISTS || exportOrder(env, markerNames, &map))
        exportMap(env, &map, distsMoth, distsFath, errorsMoth, errorsFath);
    
    freeMap(&map);
}

JNIEXPORT void JNICALL
Java_com_myriad_tmap_compute_Pedigree_writeMapNative(JNIEnv *env,
                                                       jobject this,
                                                       jlong pedPtr,
                                                       jobjectArray markerNames,
                                                       jboolean split,
                                                       jobject path)
{
    pedigree_t  *ped = (pedigree_t *)pedPtr;
    map_t       map;
    const char  *cPath;
    FILE        *fp;
    
    orderFromJStrings(env, ped, &map, markerNames, split);
    
    computeMapLlhd(&map);
    
    cPath = (*env)->GetStringUTFChars(env, path, NULL);
    fp = fopen(cPath, "w");
    (*env)->ReleaseStringUTFChars(env, path, cPath);
    if (fp) {
        normalizeMap(&map);
        printMap(fp, &map);
        fclose(fp);
    } else
        fprintf(stderr, "Couldn't open %s\n", cPath);
}

/* Takes a path in a jstring and reads the pedigree file indicated by that
   path. */
int readPedFileJstring(JNIEnv *env, pedigree_t *ped, jstring path)
{
    const char  *cPath = (*env)->GetStringUTFChars(env, path, NULL);
    int         success;
    
    success = readFile(ped, cPath);
    (*env)->ReleaseStringUTFChars(env, path, cPath);
    
    return success;
}

/* Initializes the map structure to have the marker order indicated by the
   array of Java strings markerNames.
 */
void orderFromJStrings(JNIEnv *env, pedigree_t *ped, map_t *map,
                       jobjectArray markerNames, jboolean split)
{
    int     numMarks = (*env)->GetArrayLength(env, markerNames);
    int     i, j;
    
    allocMap(map, ped, numMarks, split);
    
    for (i = 0; i < numMarks; ++i) {
        jobject     name = (*env)->GetObjectArrayElement(env, markerNames, i);
        const char  *cName = (*env)->GetStringUTFChars(env, name, NULL);
        
        for (j = 0; j < ped->numMarks; ++j) {
            if (strcmp(cName, ped->markers[j].name) == 0) {
                map->order[i] = ped->markers + j;
                break;
            }
        }
        if (j >= ped->numMarks) {
            fprintf(stderr, "Couldn't find marker %s\n", cName);
            map->order[i] = ped->markers;
        }
        
        (*env)->ReleaseStringUTFChars(env, name, cName);
    }
}

/* Exports the quantitative components of the map into the errors array
   (must already be allocated) and the return value
 */
void exportMap(JNIEnv *env, map_t *map, jdoubleArray distsMoth,
               jdoubleArray distsFath, jdoubleArray errorsMoth,
               jdoubleArray errorsFath)
{
    double          *dists;
    int             i;
    
    (*env)->SetDoubleArrayRegion(env, errorsMoth, 0, map->numMarks,
                                 map->errorsMoth);
    (*env)->SetDoubleArrayRegion(env, errorsFath, 0, map->numMarks,
                                 map->errorsFath);
    
    if (map->numMarks <= 1)
        return;
    
    dists = (double *)malloc((map->numMarks - 1) * sizeof(double));
    
    for (i = 0; i < map->numMarks - 1; ++i)
        dists[i] = theta2Dist(map->recombsMoth[i]);
    (*env)->SetDoubleArrayRegion(env, distsMoth, 0, map->numMarks - 1, dists);
    
    for (i = 0; i < map->numMarks - 1; ++i)
        dists[i] = theta2Dist(map->recombsFath[i]);
    (*env)->SetDoubleArrayRegion(env, distsFath, 0, map->numMarks - 1, dists);
    
    free(dists);
}

/* Exports the marker order from map into the array markerNames, which must be
   already allocated. Returns FALSE if there has been an error.
 */
int exportOrder(JNIEnv *env, jobjectArray markerNames, map_t *map)
{
    int     i;
    
    for (i = 0; i < map->numMarks; ++i) {
        jstring     str = (*env)->NewStringUTF(env, map->order[i]->name);
        
        (*env)->SetObjectArrayElement(env, markerNames, i, str);
    }
    
    return TRUE;
}
