/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "map.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* used by normalizeMap to indicate insiginficant levels of recombination. */
#define RECOMB_INSIGNIF 1.0e-5
#define MAX_DIST 300.0

static void swapRecombError(map_t *map, int recombNdx, int errorNdx);
static void normalizeMapEnds(map_t *map);
static void normalizeMapSplit(map_t *map);
static void normalizeMapSplitEnd(map_t *map, int start, int dir, int recombOff,
                                 int end);
static void normalizeMapSplitErrors(map_t *map);

void allocMap(map_t *map, pedigree_t *ped, int numMarks, int separateParents)
{
    map->ped = ped;
    map->numMarks = numMarks;
    
    map->order = (marker_t **)malloc(numMarks * sizeof(marker_t *));
    map->recombsMoth = (prob_t *)malloc((numMarks - 1) * sizeof(prob_t));
    if (separateParents)
        map->recombsFath = (prob_t *)malloc((numMarks - 1) * sizeof(prob_t));
    else
        map->recombsFath = map->recombsMoth;
    map->errorsMoth = (prob_t *)malloc(numMarks * sizeof(prob_t));
    if (separateParents)
        map->errorsFath = (prob_t *)malloc(numMarks * sizeof(prob_t));
    else
        map->errorsFath = map->errorsMoth;
    map->logLlhd = 0.0;
}

/* Allocates a new map which has the same ped, size and separateness as the
   model.
 */
void allocSimilarMap(map_t *map, const map_t *model)
{
    int     numMarks = model->numMarks;
    
    map->ped = model->ped;
    map->numMarks = numMarks;
    
    map->order = (marker_t **)malloc(numMarks * sizeof(marker_t *));
    map->recombsMoth = (prob_t *)malloc((numMarks - 1) * sizeof(prob_t));
    if (model->recombsMoth != model->recombsFath)
        map->recombsFath = (prob_t *)malloc((numMarks - 1) * sizeof(prob_t));
    else
        map->recombsFath = map->recombsMoth;
    map->errorsMoth = (prob_t *)malloc(numMarks * sizeof(prob_t));
    if (model->errorsMoth != model->errorsFath)
        map->errorsFath = (prob_t *)malloc(numMarks * sizeof(prob_t));
    else
        map->errorsFath = map->errorsMoth;
    map->logLlhd = 0.0;
}

void freeMap(map_t *map)
{
    free(map->order);
    free(map->recombsMoth);
    if (map->recombsMoth != map->recombsFath)
        free(map->recombsFath);
    free(map->errorsMoth);
    if (map->errorsMoth != map->errorsFath)
        free(map->errorsFath);
}

/* Caller is responsible for ensuring that dst has sufficient space allocated
   and that dst has the same separateness for the recombs and errors. If src
   separates the parents for either errors or recombs, then dst must do the
   same.
 */
void copyMap(map_t *dst, const map_t *src)
{
    dst->ped = src->ped;
    dst->numMarks = src->numMarks;
    dst->logLlhd = src->logLlhd;
    
    memcpy(dst->order, src->order, src->numMarks * sizeof(marker_t *));
    memcpy(dst->recombsMoth, src->recombsMoth,
           (src->numMarks - 1) * sizeof(prob_t));
    memcpy(dst->recombsFath, src->recombsFath,
           (src->numMarks - 1) * sizeof(prob_t));
    memcpy(dst->errorsMoth, src->errorsMoth, src->numMarks * sizeof(prob_t));
    memcpy(dst->errorsFath, src->errorsFath, src->numMarks * sizeof(prob_t));
}

void swapMaps(map_t *a, map_t *b)
{
    pedigree_t  *ped;
    int         numMarks;
    marker_t    **order;
    prob_t      *recombs;
    prob_t      logLlhd;
    
    ped = a->ped;
    a->ped = b->ped;
    b->ped = ped;

    numMarks = a->numMarks;
    a->numMarks = b->numMarks;
    b->numMarks = numMarks;
    
    order = a->order;
    a->order = b->order;
    b->order = order;
    
    recombs = a->recombsMoth;
    a->recombsMoth = b->recombsMoth;
    b->recombsMoth = recombs;
    
    recombs = a->recombsFath;
    a->recombsFath = b->recombsFath;
    b->recombsFath = recombs;
    
    recombs = a->errorsMoth;
    a->errorsMoth = b->errorsMoth;
    b->errorsMoth = recombs;
    
    recombs = a->errorsFath;
    a->errorsFath = b->errorsFath;
    b->errorsFath = recombs;
    
    logLlhd = a->logLlhd;
    a->logLlhd = b->logLlhd;
    b->logLlhd = logLlhd;
}

prob_t theta2Dist(prob_t theta)
{
    return theta2DistKosambi(theta);
}

/* Computes the Haldane distance for a given recombination fraction. */
prob_t theta2DistHaldane(prob_t theta)
{
    prob_t  dist = -log(1 - 2.0 * theta) * 50.0;
    
    if (theta >= 0.5 || dist > MAX_DIST)
        return MAX_DIST;
    else if (dist > 0.0)
        return dist;
    else
        return 0.0;
}

/* Computes the Kosambi distance for a given recombination fraction. */
prob_t theta2DistKosambi(prob_t theta)
{
    prob_t  dist = log((1.0 + 2.0 * theta)/(1.0 - 2.0 * theta)) * 25.0;
    
    if (theta >= 0.5 || dist > MAX_DIST)
        return MAX_DIST;
    else if (dist > 0.0)
        return dist;
    else
        return 0.0;
}

prob_t dist2Theta(prob_t dist)
{
    return distKosambi2Theta(dist);
}

/* Maps Kosambi distance to recombination fraction */
double distKosambi2Theta(double dist)
{
    double  e = exp(dist / 25.0);
    
    return (e - 1)/(e + 1) / 2.0;
}

/* Prints the map in a format suitable for human reading. */
void printMap(FILE *fp, map_t *map)
{
    int     i;
    prob_t  totalDistMoth = 0.0, totalDistFath = 0.0;
    
    fprintf(fp, "name          \t");
    if (map->recombsMoth == map->recombsFath)
        fprintf(fp, "pos\tdist\t");
    else
        fprintf(fp, "pos m\tpos f\tdist m\tdist f\t");
    if (map->errorsMoth == map->errorsFath)
        fprintf(fp, "error\n");
    else
        fprintf(fp, "error m\terror f\n");
    
    for (i = 0; i < map->numMarks; ++i) {
        int     informFather, informMother;
        
        fprintf(fp, "%-14s\t%.1f\t", map->order[i]->name, totalDistMoth);
        if (map->recombsFath != map->recombsMoth)
            fprintf(fp, "%.1f\t", totalDistFath);
        
        if (i == map->numMarks - 1) {
            fprintf(fp, "--");
            if (map->recombsFath != map->recombsMoth)
                fprintf(fp, "\t--");
        } else {
            prob_t  distMoth = theta2Dist(map->recombsMoth[i]);
            prob_t  distFath = theta2Dist(map->recombsFath[i]);
            fprintf(fp, "%.1f", distMoth);
            if (map->recombsFath != map->recombsMoth)
                fprintf(fp, "\t%.1f", distFath);
            totalDistMoth += distMoth;
            totalDistFath += distFath;
        }
        
        if (map->errorsMoth != map->errorsFath)
            fprintf(fp, "\t%.4f", map->errorsMoth[i]);
        markerWhereInform(map->order[i], map->ped->numInds, &informMother,
                          &informFather);
        fprintf(fp, "\t%.4f\t%c%c\n", map->errorsFath[i],
                informMother ? 'm' : ' ', informFather ? 'f' : ' ');
    }
}

/* There are certain classes of maps which will all have identical log
   likelihoods. This function picks between these equivalent maps on the basis
   of criteria not accounted for in the likelihood calculation. It is equivalent
   to having various infintessimally small terms in the likelihood calculation.
 */
void normalizeMap(map_t *map)
{
    /* Can't normalize very small maps. */
    if (map->numMarks <= 2)
        return;
    
    normalizeMapDists(map);
    
    /* if errors and recombs aren't both split or both not split, then we can't
       normalize. */
    if (map->recombsMoth != map->recombsFath) {
        if (map->errorsMoth != map->errorsFath)
            normalizeMapEnds(map);
    } else {
        if (map->errorsMoth == map->errorsFath)
            normalizeMapEnds(map);
    }
}

/* This is the similar to normalizeMap, but only performs the normalizations
   which do not affect the marker order.
 */
void normalizeMapDists(map_t *map)
{
    /* Can't normalize very small maps. */
    if (map->numMarks <= 2)
        return;
    
    if (map->errorsMoth != map->errorsFath)
        normalizeMapSplitErrors(map);
    
    if (map->recombsMoth != map->recombsFath)
        normalizeMapSplit(map);
}

/* The order of the 2 markers at the end is indeterminate, so this function
   picks the one which minimizes the error. This only applies if the errors and
   the the recombinations are both split or both not split.
 */
void normalizeMapEnds(map_t *map)
{
    marker_t    *m;
    int         i;
    int         ndx;
    prob_t      maxRecomb;
    
    ndx = 0;
    maxRecomb = map->recombsMoth[0] + map->recombsFath[0];
    
    for (i = 1; i < map->numMarks - 1; ++i) {
        prob_t  error = map->errorsMoth[i] + map->errorsFath[i];
        
        if (error > maxRecomb) {
            maxRecomb = error;
            ndx = i;
        }
        if (map->recombsMoth[i] + map->recombsFath[i] >= RECOMB_INSIGNIF)
            break;
    }
    
    if (ndx != 0) {
        swapRecombError(map, 0, ndx);
        
        m = map->order[0];
        map->order[0] = map->order[ndx];
        map->order[ndx] = m;
    }
    
    ndx = map->numMarks - 1;
    maxRecomb = map->recombsMoth[map->numMarks - 2]
                    + map->recombsFath[map->numMarks - 2];
    
    for (i = map->numMarks - 2; i > 0; --i) {
        prob_t  error = map->errorsMoth[i] + map->errorsFath[i];
        
        if (error > maxRecomb) {
            maxRecomb = error;
            ndx = i;
        }
        
        if (map->recombsMoth[i - 1] + map->recombsFath[i] >= RECOMB_INSIGNIF)
            break;
    } 
    
    if (ndx != map->numMarks - 1) {
        swapRecombError(map, map->numMarks - 2, ndx);
        
        m = map->order[ndx];
        map->order[ndx] = map->order[map->numMarks - 1];
        map->order[map->numMarks - 1] = m;
    }
}

void swapRecombError(map_t *map, int i, int j)
{
    prob_t  tmpMoth = map->errorsMoth[j];
    prob_t  tmpFath = map->errorsFath[j];
    
    map->errorsMoth[j] = map->recombsMoth[i];
    map->errorsFath[j] = map->recombsFath[i];
    map->recombsMoth[i] = tmpMoth;
    map->recombsFath[i] = tmpFath;
}

/* This function only applies to maps with split recombination fractions.   
   
   If there are one or markers at either end which are only informative in one
   parent, then the distances for the other parent will be completely
   undetermined. For convenience, this function sets them to be the same as the
   other parent. This is similar to having an infintessimal penalty proportional
   to the disagreement between the two distances.
 */
void normalizeMapSplit(map_t *map)
{
    normalizeMapSplitEnd(map, 0, 1, 0, map->numMarks - 2);
    normalizeMapSplitEnd(map, map->numMarks - 1, -1, -1, 0);
}

void normalizeMapSplitEnd(map_t *map, int start, int dir, int recombOff,
                          int end)
{
    int     i;
    int     informMoth, informFath;
    int     *watch;
    prob_t  *undefRecombs, *recombs;
    
    markerWhereInform(map->order[start], map->ped->numInds, &informMoth,
                      &informFath);
    if (!informMoth && informFath) {
        watch = &informMoth;
        undefRecombs = map->recombsMoth;
        recombs = map->recombsFath;
    } else if (informMoth && !informFath) {
        watch = &informFath;
        undefRecombs = map->recombsFath;
        recombs = map->recombsMoth;
    } else
        return;
    
    for (i = start; i != end && !*watch; i += dir) {
        undefRecombs[i + recombOff] = recombs[i + recombOff];
        markerWhereInform(map->order[i + dir], map->ped->numInds, &informMoth,
                          &informFath);
    }
    
}

/* This function works on maps with split errors. If a marker is uninformative
   for one parent, then the error for that parent is also undefined. This
   function sets the two errors to be equal in that case.
 */
void normalizeMapSplitErrors(map_t *map)
{
    int     i;
    
    for (i = 0; i < map->numMarks; ++i) {
        int     moth, fath;
        
        markerWhereInform(map->order[i], map->ped->numInds, &moth, &fath);
        if (moth && !fath)
            map->errorsFath[i] = map->errorsMoth[i];
        else if (!moth && fath)
            map->errorsMoth[i] = map->errorsFath[i];
    }
}
