/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "mapLlhd.h"
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#define OFF_MAP_DISTS 1

#define INITIAL_DISTANCE 0.10
#if OFF_MAP_DISTS
#define INITIAL_OFF_MAP_DISTANCE INITIAL_DISTANCE
#else
#define INITIAL_OFF_MAP_DISTANCE 0.0
#endif
    /* We consider convergence to have been achieved when successive log
       likelihoods differ by less than this amount. */
#define LOG_LIKELIHOOD_EPS 0.001
#define RECOMB_EPS 0.001
    /* Recombination fractions less than this are treated as zero. */
#define RECOMB_INSIGNIF 1.0e-7

typedef prob_t genProbs_t[NUM_GENOTYPES];

static void computeMapLlhdWithInits(map_t *map);
static void emStep(map_t *map);
static void computeMarkerProbs(const map_t *map, genProbs_t *markerProbs);
static void computeCondProbs(const map_t *map, genProbs_t *probs,
                             genProbs_t *probsPreMarker,
                             genProbs_t *markerProbs, int dir, prob_t *);
#if OFF_MAP_DISTS
static void computeOffMapDists(map_t *map, genProbs_t *leftPreProbs,
                               genProbs_t *rightPreProbs);
#endif
static void computeDist(genProbs_t *probs1, genProbs_t *probs2, int numInds,
                        prob_t *recombMoth, prob_t *recombFath);
static void computeProbsRecomb(genProbs_t probs, prob_t recombMoth,
                               prob_t recombFath);

/* Computes a likelihood similar to that in linkage analysis. In particular, it
   fixes all the parameters of the map except the error rate of the marker with
   index traitMarker in the map and then finds the maximum likelihood error
   rate for that marker and the corresponding log likelihood of the map.
   
   The recombs* and errors* arrays of map must be initialized to reasonable
   values, with the possible exception of the errors for the marker
   traitMarker.
   
   This works well enough if the "trait" is an actual marker. It doesn't work
   so well with actual traits because:
     - The trait cannot be quantitative
     - The trait must be phased correctly
     - It assumes complete penetrance. While the error provides compensation
       for this, it is not as general as a true model for the trait.
 */
void quasiLinkageAnalysis(map_t *map, int traitMarker)
{
    genProbs_t  *markerProbs;
    genProbs_t  *unusedProbs, *leftPreProbs, *rightPreProbs;
    genProbs_t  *locusProbs; /* likelihoods of genotypes at trait position,
                                ignoring trait marker */
    genProbs_t  *traitProbs; /* likelihoods of genotypes for trait marker */
    int         numInds = map->ped->numInds;
    int         numObs = map->numMarks * numInds;
    double      recombDiff;
    int         i, j;
    
    markerProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    unusedProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    leftPreProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    rightPreProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    locusProbs = (genProbs_t *)malloc(numInds * sizeof(genProbs_t));
    traitProbs = (genProbs_t *)malloc(numInds * sizeof(genProbs_t));
    
    map->errorsFath[traitMarker] = map->errorsMoth[traitMarker]
        = INITIAL_OFF_MAP_DISTANCE;
    
    computeMarkerProbs(map, markerProbs);
    computeCondProbs(map, unusedProbs, leftPreProbs, markerProbs, 1, NULL);
    computeCondProbs(map, unusedProbs, rightPreProbs, markerProbs, -1, NULL);
    
    for (i = 0; i < numInds; ++i) {
        int     ndx = traitMarker * numInds + i;
        
        for (j = 0; j < NUM_GENOTYPES; ++j) {
            traitProbs[i][j]
                = map->order[traitMarker]->obs[i] & (1 << j) ? 1.0 : 0.0;
            locusProbs[i][j] = rightPreProbs[ndx][j] * leftPreProbs[ndx][j];
        }
    }
    
    /* EM algorithm to compute the error rate of the trait marker. */
    do {
        double  oldRecomb = map->errorsMoth[traitMarker];
        computeDist(locusProbs, traitProbs, numInds,
                    map->errorsMoth + traitMarker,
                    map->errorsFath + traitMarker);
        recombDiff = map->errorsMoth[traitMarker] - oldRecomb;
    } while (recombDiff > RECOMB_EPS || recombDiff < -RECOMB_EPS);
    
    /* This is necessary to compute the log likelihood of the map. */
    computeMarkerProbs(map, markerProbs);
    computeCondProbs(map, unusedProbs, leftPreProbs, markerProbs, 1,
                     &map->logLlhd);
    
    free(markerProbs);
    free(unusedProbs);
    free(leftPreProbs);
    free(rightPreProbs);
    free(locusProbs);
    free(traitProbs);
}

/* Computes the maximum likelihood map for a given marker order. The distances,
   error rates, and log-10 likelihood are stored in the map structure.
 */
void computeMapLlhd(map_t *map)
{
    int     i;
    
    if (map->numMarks == 0)
        return;
    
    for (i = 0; i < map->numMarks - 1; ++i) {
        map->recombsFath[i] = map->recombsMoth[i] = INITIAL_DISTANCE;
        map->errorsFath[i] = map->errorsMoth[i] = INITIAL_OFF_MAP_DISTANCE;
    }
    map->errorsFath[i] = map->errorsFath[0]
        = map->errorsMoth[i] = map->errorsMoth[0] = 0.0;
    
    computeMapLlhdWithInits(map);
}

/* Computes the maximum likelihood map, assuming that the distances and errors
   have already been assigned reasonable seed values.
 */
void computeMapLlhdWithInits(map_t *map)
{
    prob_t  oldLogLlhd;
    
    emStep(map);
    do {
        oldLogLlhd = map->logLlhd;
        emStep(map);
    } while(map->logLlhd - oldLogLlhd > LOG_LIKELIHOOD_EPS);
}

/* Performs one step of the EM-algorithm. It computes new recombination
   fractions which have a likelihood no less than the old ones. */
void emStep(map_t *map)
{
    genProbs_t  *markerProbs;
    genProbs_t  *leftProbs, *leftPreProbs;
    genProbs_t  *rightProbs, *rightPreProbs;
    int         i;
    int         numInds = map->ped->numInds;
    int         numObs = map->numMarks * numInds;
    
    markerProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    leftProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    leftPreProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    rightProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    rightPreProbs = (genProbs_t *)malloc(numObs * sizeof(genProbs_t));
    
    computeMarkerProbs(map, markerProbs);
    computeCondProbs(map, leftProbs, leftPreProbs, markerProbs, 1,
                     &map->logLlhd);
    computeCondProbs(map, rightProbs, rightPreProbs, markerProbs, -1, NULL);
    
    for (i = 0; i < map->numMarks - 1; ++i)
        computeDist(rightProbs + (i + 1) * numInds, leftProbs + i * numInds,
                    numInds, map->recombsMoth + i, map->recombsFath + i);
    
#if OFF_MAP_DISTS
    computeOffMapDists(map, leftPreProbs, rightPreProbs);
#endif
    
    free(markerProbs);
    free(leftProbs);
    free(leftPreProbs);
    free(rightProbs);
    free(rightPreProbs);
}

/* Computes the probability distributions of the different genotypes at each
   of the loci. */
void computeMarkerProbs(const map_t *map, genProbs_t *markerProbs)
{
    int     i, j, k, l;
    int     numInds = map->ped->numInds;
    
    for (i = 0; i < map->numMarks; ++i) {
        prob_t      probRecombMoth = map->errorsMoth[i];
        prob_t      probRecombFath = map->errorsFath[i];
        genProbs_t  probRecombs;
        
        computeProbsRecomb(probRecombs, probRecombMoth, probRecombFath);
        
        for (j = 0; j < numInds; ++j) {
            prob_t      *probs = markerProbs[i * numInds + j];
            obs_t       obs = map->order[i]->obs[j];
            
            if (probRecombMoth < RECOMB_INSIGNIF
                    && probRecombFath < RECOMB_INSIGNIF) {
                for (k = 0; k < NUM_GENOTYPES; ++k)
                    probs[k] = (obs & (1 << k)) ? 1.0 : 0.0;
            } else {
                for (l = 0; l < NUM_GENOTYPES; ++l)
                    probs[l] = 0.0;
                for (k = 0; k < NUM_GENOTYPES; ++k) {
                    if (obs & (1 << k)) {
                        for (l = 0; l < NUM_GENOTYPES; ++l)
                            probs[l] += probRecombs[k ^ l];
                    }
                }
            }
        }
    }
}

void computeCondProbs(const map_t *map, genProbs_t *probs,
                      genProbs_t *probsPreMarker,
                      genProbs_t *markerProbs, int dir, prob_t *logLlhd)
{
    int     numInds = map->ped->numInds;
    int     start = dir > 0 ? 0 : map->numMarks - 1;
    int     end = start + map->numMarks * dir;
    int     distVecOff = dir > 0 ? -1 : 0; /* offset into the distances vector*/
    int     i, j, gen, lastGen;
    prob_t  lnLlhd = 0.0;
    
    for (i = start * numInds; i != (start + 1) * numInds; ++i) {
        for (j = 0; j < NUM_GENOTYPES; ++j) {
            probsPreMarker[i][j] = 1.0;
            probs[i][j] = markerProbs[i][j] / NUM_GENOTYPES;
        }
    }
    
    for (i = start + dir; i != end; i += dir) {
        prob_t  probRecombMoth = map->recombsMoth[i + distVecOff];
        prob_t  probRecombFath = map->recombsFath[i + distVecOff];
        prob_t  probRecombs[NUM_GENOTYPES];
        
        computeProbsRecomb(probRecombs, probRecombMoth, probRecombFath);
        
        for (j = i * numInds; j < (i + 1) * numInds; ++j) {
            int         prev = j - dir * numInds;
            genProbs_t  a;
            prob_t      b = 0.0;
            
            for (gen = 0; gen < NUM_GENOTYPES; ++gen) {
                if (probRecombMoth < RECOMB_INSIGNIF
                        && probRecombFath < RECOMB_INSIGNIF)
                    probsPreMarker[j][gen] = probs[prev][gen];
                else {
                    probsPreMarker[j][gen] = 0.0;
                    
                    for (lastGen = 0; lastGen < NUM_GENOTYPES; ++lastGen)
                        probsPreMarker[j][gen] += probRecombs[gen ^ lastGen]
                                                    * probs[prev][lastGen];
                }
                
                a[gen] = probsPreMarker[j][gen] * markerProbs[j][gen];
                b += a[gen];
            }
            
            for (gen = 0; gen < NUM_GENOTYPES; ++gen)
                probs[j][gen] = a[gen] / b;
            if (logLlhd)
                lnLlhd += log(b);
        }
    }
    if (logLlhd)
        *logLlhd = lnLlhd * M_LOG10E;
}

#if OFF_MAP_DISTS
void computeOffMapDists(map_t *map, genProbs_t *leftPreProbs,
                        genProbs_t *rightPreProbs)
{
    genProbs_t      *locusProbs, *markerProbs;
    int             i, j, k;
    int             numInds = map->ped->numInds;
    
    locusProbs = (genProbs_t *)malloc(numInds * sizeof(genProbs_t));
    markerProbs = (genProbs_t *)malloc(numInds * sizeof(genProbs_t));
    
    for (i = 1; i < map->numMarks - 1; ++i) {
        for (j = 0; j < numInds; ++j) {
            int     ndx = i * numInds + j;
            for (k = 0; k < NUM_GENOTYPES; ++k) {
                locusProbs[j][k] = rightPreProbs[ndx][k] * leftPreProbs[ndx][k];
                if (map->order[i]->obs[j] & (1 << k))
                    markerProbs[j][k] = 1.0;
                else
                    markerProbs[j][k] = 0.0;
            }
        }
        
        computeDist(locusProbs, markerProbs, numInds, map->errorsMoth + i,
                    map->errorsFath + i);
    }
    
    free(locusProbs);
    free(markerProbs);
}
#endif

void computeProbsRecomb(genProbs_t probs, prob_t recombMoth, prob_t recombFath)
{
    probs[0] = (1.0 - recombMoth) * (1.0 - recombFath);
    probs[1] = recombMoth * (1.0 - recombFath);
    probs[2] = (1.0 - recombMoth) * recombFath;
    probs[3] = recombMoth * recombFath;
}

void computeDist(genProbs_t *probs1, genProbs_t *probs2, int numInds,
                   prob_t *distMoth, prob_t *distFath)
{
    prob_t          recombsMoth = 0.0, recombsFath = 0.0;
    genProbs_t      probRecombs;
    int             i, j, k;
    
    if (*distMoth < RECOMB_INSIGNIF && *distFath < RECOMB_INSIGNIF) {
        *distMoth = *distFath = 0.0;
        return;
    }
    
    computeProbsRecomb(probRecombs, *distMoth, *distFath);
    
    for (i = 0; i < numInds; ++i) {
        prob_t      aMoth = 0.0, aFath = 0.0, b = 0.0;
        
        for (j = 0; j < NUM_GENOTYPES; ++j) {
            prob_t  cMoth = 0.0, cFath = 0.0, d = 0.0;
            
            for (k = 0; k < NUM_GENOTYPES; ++k) {
                prob_t      e = probs2[i][k] * probRecombs[j ^ k];
                if ((j ^ k) & 0x1) cMoth += e;
                if ((j ^ k) & 0x2) cFath += e;
                d += e;
            }
            
            aMoth += cMoth * probs1[i][j];
            aFath += cFath * probs1[i][j];
            b += d * probs1[i][j];
        }
        
        recombsMoth += aMoth / b;
        recombsFath += aFath / b;
    }
    
    if (distMoth == distFath)
        *distMoth = (recombsMoth + recombsFath) / numInds / 2.0;
    else {
        *distMoth = recombsMoth / numInds;
        *distFath = recombsFath / numInds;
    }
}
