/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "genmap.h"
#include "ped.h"
#include "input.h"
#include "inputOrder.h"
#include "mapLlhd.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define STEP_SIZE 0.1

/* Takes a single marker out of a linkage group and treats it like a phenotype
   in a linkage analysis. It outputs the log likelihood as a function of
   position to standard out.
 */
int main(int argc, char **argv)
{
    pedigree_t  ped;
    marker_t    **markers;
    int         numMarks;
    map_t       map;
    map_t       refMap; /* map of the markers without the trait marker */
    marker_t    *movingMarker = NULL;
    prob_t      maxLogLlhd = 0.0;
    int         i, j;
    prob_t      pos; /* distance past the first marker */
    prob_t      intervalSize = 0.0;
    int         intervalFlag = FALSE;
    prob_t      intervalStepDown = 0.0;
    prob_t      totalSize = 0.0;
    prob_t      *llhds;
    
    if (argc == 6 && strcmp(argv[1], "-i") == 0) {
        char    *endptr;
        
        intervalStepDown = strtod(argv[2], &endptr);
        if (*endptr != '\0') {
            fprintf(stderr, "Error: %s: %s is not a valid number\n", *argv,
                    argv[2]);
            exit(1);
        }
        intervalFlag = TRUE;
    } else if (argc != 4 || argv[1][0] == '-') {
        fprintf(stderr, "Usage: %s <pedigree> <order> <marker name>\n", *argv);
        return EXIT_FAILURE;
    }
    
    if (!readFile(&ped, argv[argc - 3]))
        return EXIT_FAILURE;
    if ((markers = readOrderFile(&ped, &numMarks, argv[argc - 2])) == NULL)
        return EXIT_FAILURE;
    
    for (i = 0; i < ped.numMarks; ++i) {
        if (strcmp(ped.markers[i].name, argv[argc - 1]) == 0) {
            movingMarker = ped.markers + i;
            break;
        }
    }
    
    if (movingMarker == NULL) {
        fprintf(stderr, "Marker %s not found in order\n", argv[argc - 1]);
        return EXIT_FAILURE;
    }
    
    for (i = 0; i < numMarks; ++i) {
        if (markers[i] == movingMarker) {
            --numMarks;
            break;
        }
    }
    
    /* At this point numMarks is the number of markers in the linkage group,
       not counting the trait marker. */
    
    /* Create map of markers without the trait marker. */
    allocMap(&refMap, &ped, numMarks, FALSE);
    memcpy(refMap.order, markers, i * sizeof(marker_t *));
    memcpy(refMap.order + i, markers + i + 1,
           (numMarks - i) * sizeof(marker_t *));
    computeMapLlhd(&refMap);
    
    /* Set up map with parameters copied from refMap and an empty slot at
       position 0 for the trait marker. */
    allocMap(&map, &ped, numMarks + 1, FALSE);
    memcpy(map.order + 1, refMap.order, numMarks * sizeof(marker_t *));
    memcpy(map.recombsMoth + 1, refMap.recombsMoth,
           (numMarks - 1) * sizeof(prob_t));
    memcpy(map.errorsMoth + 1, refMap.errorsMoth,
           numMarks * sizeof(prob_t));
    
    totalSize = 0.0;
    for (i = 0; i < numMarks - 1; ++i)
        totalSize += theta2Dist(refMap.recombsMoth[i]);
    llhds = (prob_t *)malloc((1 + (int) (totalSize / STEP_SIZE))
                                * sizeof(prob_t));
    
    for (i = 1, pos = 0.0, j = 0; i < numMarks; ++i) {
        /* place the trait marker at position i, meaning between markers i-1
           and i from the reference sequence. */
        prob_t  curGapSz = theta2Dist(refMap.recombsMoth[i - 1]);
        
        map.order[i - 1] = map.order[i];
        map.order[i] = movingMarker;
        map.errorsMoth[i - 1] = map.errorsMoth[i];
        
        for (; pos < curGapSz; pos += STEP_SIZE, ++j) {
            map.recombsMoth[i - 1] = dist2Theta(pos);
            map.recombsMoth[i] = dist2Theta(curGapSz - pos);
            
            quasiLinkageAnalysis(&map, i);
            
            llhds[j] = map.logLlhd;
            if (j == 0 || map.logLlhd > maxLogLlhd)
                maxLogLlhd = map.logLlhd;
        }
        
        pos -= curGapSz;
        map.recombsMoth[i - 1] = refMap.recombsMoth[i - 1];
    }
    
    for (i = 0; i < j; ++i) {
        if (intervalFlag) {
            if (maxLogLlhd - llhds[i] < intervalStepDown)
                intervalSize += STEP_SIZE;
        } else
            printf("%.2f %.2f %.2f\n", i * STEP_SIZE, llhds[i],
                   llhds[i] - maxLogLlhd);
    }
    
    if (intervalFlag)
        printf("%.2f\n", intervalSize);
    
    free(llhds);
    freeMap(&refMap);
    freeMap(&map);
    free(markers);
    freePed(&ped);
    
    return EXIT_SUCCESS;
}
